import Foundation

/// A type-safe wrapper around UserDefaults that supports JSON-based Codable encoding/decoding.
///
/// `TypedUserDefaults` mimics Apple's UserDefaults API while providing type safety for complex data structures
/// that conform to ``TypedAppStorageValue``. Since ``TypedAppStorageValue`` includes the storage key and default value,
/// many parameters become optional in the API.
public class TypedUserDefaults {
  /// The underlying UserDefaults store.
  public let userDefaults: UserDefaults
  private let encoder = JSONEncoder()
  private let decoder = JSONDecoder()

  /// The shared typed user defaults object.
  public static let standard = TypedUserDefaults()

  /// Creates a TypedUserDefaults instance with the specified UserDefaults store.
  /// - Parameter userDefaults: The underlying UserDefaults store. Defaults to UserDefaults.standard.
  public init(userDefaults: UserDefaults = .standard) {
    self.userDefaults = userDefaults
  }

  // MARK: - Reading Values

  /// Returns the typed object associated with the specified key.
  /// - Parameters:
  ///   - type: The type of object to retrieve.
  ///   - key: The key to retrieve the value for. If nil, uses the type's appStorageKey.
  ///   - defaultValue: The default value to return if no object exists for the key. If nil, uses the type's default value.
  /// - Returns: The typed object, or the type's default value if no object exists for the key.
  public func object<T: TypedAppStorageValue>(
    of type: T.Type,
    forKey key: String? = nil,
    defaultValue: T? = nil
  ) -> T {
    let storageKey = key ?? T.appStorageKey

    guard let jsonString = userDefaults.string(forKey: storageKey),
          let data = jsonString.data(using: .utf8),
          let decodedValue = try? decoder.decode(T.self, from: data)
    else {
      return defaultValue ?? T.defaultValue
    }

    return decodedValue
  }

  // MARK: - Writing Values

  /// Sets the value of the specified typed object for the given key.
  /// - Parameters:
  ///   - value: The typed object to store.
  ///   - key: The key to store the value under. If nil, uses the type's appStorageKey.
  public func set<T: TypedAppStorageValue>(
    _ value: T,
    forKey key: String? = nil
  ) {
    let storageKey = key ?? T.appStorageKey

    guard let data = try? encoder.encode(value),
          let jsonString = String(data: data, encoding: .utf8)
    else {
      return
    }

    userDefaults.set(jsonString, forKey: storageKey)
  }

  // MARK: - Removing Values

  /// Removes the value associated with the specified key for the given type.
  /// - Parameters:
  ///   - type: The type of object to remove.
  ///   - key: The key to remove the value for. If nil, uses the type's appStorageKey.
  public func removeObject<T: TypedAppStorageValue>(
    of type: T.Type,
    forKey key: String? = nil
  ) {
    let storageKey = key ?? T.appStorageKey
    userDefaults.removeObject(forKey: storageKey)
  }

  // MARK: - UserDefaults Forwarding

  /// Synchronizes any changes made to the shared user defaults.
  /// - Returns: true if the synchronization was successful, false otherwise.
  @discardableResult
  public func synchronize() -> Bool {
    return userDefaults.synchronize()
  }
}
