import SwiftUI
import Testing

@testable import TypedAppStorage

struct PreferredFruit: TypedAppStorageValue, Equatable {
  enum Fruit: Codable {
    case apple, pear, banana
  }
  enum Freshness: Codable {
    case veryFresh, moderate, somewhatStale
  }

  static var appStorageKey = "preferredFruit"
  static var defaultValue = PreferredFruit(.veryFresh, .apple)

  var fruit: Fruit
  var freshness: Freshness

  init(_ freshness: Freshness, _ fruit: Fruit) {
    self.fruit = fruit
    self.freshness = freshness
  }
}

struct TestArticle: View {
  @TypedAppStorage var preferredFruit: PreferredFruit

  func changePreferred(to newValue: PreferredFruit) {
    preferredFruit = newValue
  }

  var body: some View {
    Text("Test")
  }
}

struct TestArticleWithADifferentDefault: View {
  @TypedAppStorage var preferredFruit: PreferredFruit = .init(.moderate, .pear)

  func changePreferred(to newValue: PreferredFruit) {
    preferredFruit = newValue
  }

  var body: some View {
    Text("Test")
  }
}

@Suite("TypedAppStorage Tests", .serialized)
struct TypedAppStorageTests {

  init() {
    UserDefaults.standard.removeObject(forKey: "preferredFruit")
  }

  @Test("Read default value")
  func readDefaultValue() throws {
    UserDefaults.standard.removeObject(forKey: "preferredFruit")
    let testArticle = TestArticle()

    #expect(
      testArticle.preferredFruit == PreferredFruit(.veryFresh, .apple)
    )
  }

  @Test("Call-site default")
  func callSiteDefault() throws {
    UserDefaults.standard.removeObject(forKey: "preferredFruit")
    let testArticle = TestArticleWithADifferentDefault()

    #expect(testArticle.preferredFruit == .init(.moderate, .pear))
  }

  @Test("Save and read back")
  func saveAndReadBack() throws {
    UserDefaults.standard.removeObject(forKey: "preferredFruit")
    let testArticle = TestArticle()

    testArticle.changePreferred(to: .init(.somewhatStale, .banana))

    #expect(testArticle.preferredFruit == .init(.somewhatStale, .banana))
  }

  @Test("Save and read elsewhere")
  func saveAndReadElsewhere() throws {
    UserDefaults.standard.removeObject(forKey: "preferredFruit")
    let articleA = TestArticle()
    let articleB = TestArticle()

    articleA.changePreferred(to: .init(.moderate, .banana))

    #expect(articleB.preferredFruit == .init(.moderate, .banana))
  }
}
